# -*- mode: python -*-
""" Front-end scripts for song detector """
import argparse
import logging
import os
import pickle
from pathlib import Path

import arf
import h5py as h5
import numpy as np
from arf import DataTypes, create_dataset, set_uuid, timestamp_to_datetime
from dateutil import parser as date_parser

from quicksong import __version__
from quicksong.core import IntervalFinder

script_name = "quicksong"

_example_config = dict(
    spectrogram={
        "window_ms": 20.0,
        "shift_ms": 10.0,
        "frequency_range": [0.7, 20.0],
    },
    features=["total_power", "wiener_entropy"],
    target_label="x",
    smoothing={"kernel_widths": [11, 21, 51]},
    classifier={"gamma": 2, "C": 1},
    testing={"test_size": 0.4, "random_state": 45},
    intervals={"max_gap_ms": 100, "min_interval_ms": 500},
    datasets=[],
)

log = logging.getLogger("")


def valid_datetime(date_string):
    """
    Convert string to datetime object or raise ArgumentTypeError
    """
    try:
        return date_parser.parse(date_string)
    except ValueError as err:
        msg = f"'{date_string}' is not a valid date/time"
        raise argparse.ArgumentTypeError(msg) from err


def setup_log(log, debug=False):
    """Set up logging for a module.

    log: generate by calling e.g. `log = logging.getLogger("dlab.extracellular")`
    """
    ch = logging.StreamHandler()
    formatter = logging.Formatter("%(message)s")
    loglevel = logging.DEBUG if debug else logging.INFO
    log.setLevel(loglevel)
    ch.setLevel(loglevel)
    ch.setFormatter(formatter)
    log.addHandler(ch)


def all_same(seq):
    """If all elements of seq are equal, returns the value, otherwise None"""
    it = iter(seq)
    first = next(it)
    for e in it:
        if e != first:
            return None
    return first


def make_config(args):
    import yaml

    cfg = _example_config.copy()
    for lblfile in args.label_files:
        cfg["datasets"].append(lblfile.stem)
    with open(args.config, "w") as fp:
        yaml.dump(cfg, fp)
        log.info("created new configuration file '%s'", args.config)
    if args.train:
        args.dataset_dir = lblfile.parent
        args.model = None
        train_classifier(args)


def make_extractor(cfg, sampling_rate_khz):
    from quicksong import features
    from quicksong.core import FeatureExtractor, make_hanning_kernels
    from quicksong.streaming import STFT, OverlapSaveConvolver

    speccr = STFT(
        sampling_rate_khz,
        cfg["spectrogram"]["window_ms"],
        cfg["spectrogram"]["shift_ms"],
        cfg["spectrogram"]["frequency_range"],
    )
    feats = tuple(getattr(features, f) for f in cfg["features"])
    conv = OverlapSaveConvolver(make_hanning_kernels(cfg["smoothing"]["kernel_widths"]))
    return FeatureExtractor(speccr, feats, conv)


def train_classifier(args):
    import ewave
    import yaml
    from arfx import lblio

    from quicksong import core, model

    with open(args.config) as fp:
        cfg = yaml.safe_load(fp)
    log.debug(cfg)
    label_ids = {cfg["target_label"]: 0}
    log.info("Training song classifier:")
    log.info("- version: %s", __version__)

    X_all = []
    y_all = []
    for dset in cfg["datasets"]:
        log.info("- dataset: %s", dset)
        path = args.dataset_dir / dset
        with ewave.open(path.with_suffix(".wav"), "r") as fp:
            Fs = fp.sampling_rate / 1000.0
            signal = ewave.rescale(fp.read(), "f")
            log.info(" - recording: %d samples @ %.2f kHz", signal.size, Fs)
        extractor = make_extractor(cfg, Fs)
        feats = extractor.process_all(signal)
        log.info("   - calculated features %s", feats.shape)

        with open(path.with_suffix(".lbl")) as fp:
            out = []
            for lbl in lblio.read(fp):
                n = lbl["name"]
                try:
                    id = label_ids[n]
                except KeyError:
                    id = len(label_ids)
                    label_ids[n] = id
                out.append((id, lbl["start"] * 1000, lbl["stop"] * 1000))
            labels = np.array(
                out, dtype=[("id", "i4"), ("start", "f8"), ("stop", "f8")]
            )
            log.info(" - labels: %d segments", labels.size)
            X, y = core.make_training_dataset(
                feats, labels, cfg["spectrogram"]["shift_ms"]
            )
            log.info("  - labeled frames: %d", y.size)
            X_all.append(X)
            y_all.append(y)
    X = np.vstack(X_all)
    y = np.concatenate(y_all)
    assert X.shape[0] == y.size
    log.info("- training classifier...")
    classifier, score = model.train_classifier(X, y, cfg["classifier"], cfg["testing"])
    log.info("  - model score: %.3f", score)
    if args.model is None:
        args.model = args.config.with_suffix(".pkl")
    with open(args.model, "wb") as fp:
        pickle.dump(
            {"config": cfg, "classifier": classifier, "version": __version__}, fp
        )
        log.info("- wrote model to '%s'", args.model)


class ArfWriter:
    """Utility class that writes to sequentially numbered entries"""

    def __init__(self, arfp, template="song_{index:04}"):
        self.arfp = arfp
        self.entry_count = len(arfp.keys())
        self.template = template

    @property
    def filename(self):
        return self.arfp.filename

    def create_entry(self, timestamp, **attributes):
        import arf

        next_entry_name = self.template.format(index=self.entry_count)
        self.entry_count += 1
        return arf.create_entry(self.arfp, next_entry_name, timestamp, **attributes)


def extract_songs(args):
    os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
    log.info("Extracting songs:")
    log.info("- version: %s", __version__)
    with open(args.model, "rb") as fp:
        log.info("- loading classifier from '%s'", args.model)
        model = pickle.load(fp)
    # TODO check model version
    classifier = model["classifier"]
    dt = model["config"]["spectrogram"]["shift_ms"]
    finder = IntervalFinder(max_gap=args.max_gap, min_duration=args.min_duration, dt=dt)
    # TODO: more flexibility with output format
    with arf.open_file(args.output, mode="a") as ofp:
        writer = ArfWriter(ofp)
        log.info("- saving songs to '%s'", args.output)
        for fname in args.input:
            with h5.File(fname, "r") as ifp:
                log.info("- processing recordings in '%s'", ifp.filename)
                last_sample = None
                buffer = []
                intervals = []
                for entry in ifp.values():
                    if not isinstance(entry, h5.Group):
                        log.info("  ✗ %s: not an entry, skipping", entry.name)
                        continue
                    entry_time = timestamp_to_datetime(entry.attrs["timestamp"])
                    if args.ignore_before and entry_time < args.ignore_before:
                        log.info("  ✗ %s: %s is before %s, skipping", entry.name, entry_time, args.ignore_before)
                        continue
                    if args.ignore_after and entry_time > args.ignore_after:
                        log.info("  ✗ %s: %s is after %s, skipping", entry.name, entry_time, args.ignore_after)
                        continue
                    try:
                        dset = entry[args.dataset_name]
                    except KeyError:
                        log.info(
                            "  ✗ %s: does not have dataset `%s`, skipping",
                            entry.name,
                            args.dataset_name,
                        )
                        continue
                    sampling_rate = dset.attrs["sampling_rate"] / 1000.0
                    entry_start = entry.attrs["jack_frame"]
                    log.debug(
                        " - %s: start=%s, end=%s, sampling_rate=%.2f kHz",
                        entry.name,
                        entry_start,
                        entry_start + dset.size,
                        sampling_rate,
                    )
                    # is this a continuation of the previous entry?
                    if last_sample is not None and last_sample > entry_start:
                        log.debug("   - continues previous entry")
                    else:
                        # process the intervals
                        extract_intervals(
                            writer,
                            buffer,
                            intervals,
                            args.pad_before,
                            args.pad_after,
                        )
                        extractor = make_extractor(model["config"], sampling_rate)
                        finder.reset()
                        buffer = []
                        intervals = []
                    for block in extractor.process(dset):
                        pred = classifier.predict(block)
                        intervals.extend(
                            (start * dt, stop * dt)
                            for start, stop in finder.process(pred)
                        )
                    # jack_frame is a np.uint32 so it should overflow to allow
                    # comparison with next entry
                    last_sample = entry_start + dset.size
                    buffer.append(dset)


def extract_intervals(writer, dsets, intervals, pad_before, pad_after):
    import posixpath
    from datetime import timedelta

    from quicksong.core import pad_intervals

    fields_to_drop = (
        "entry_creator",
        "timestamp",
        "jack_frame",
        "jack_sampling_rate",
        "jack_usec",
        "trial_off",
    )

    if len(dsets) == 0:
        return
    dset_name = ", ".join(dset.parent.name for dset in dsets)
    if len(intervals) == 0:
        log.info("  ✗ %s: no song detected", dset_name)
        return
    sampling_rate = all_same(dset.attrs["sampling_rate"] for dset in dsets) / 1000.0
    assert sampling_rate is not None, "Entries don't have the same sampling rate!"
    parent_entry = dsets[0].parent
    entry_attrs = {
        k: v for k, v in parent_entry.attrs.items() if k not in fields_to_drop
    }
    entry_attrs.update(
        source_file=parent_entry.file.filename,
        source_uuid=b",".join(dset.parent.attrs["uuid"] for dset in dsets),
        entry_creator=f"org.meliza.quicksong/{script_name} {__version__}",
    )
    entry_attrs.pop("uuid")
    dset_attrs = dict(dsets[0].attrs)
    dset_attrs["source_uuid"] = b",".join(dset.attrs["uuid"] for dset in dsets)
    dset_attrs["datatype"] = DataTypes.ACOUSTIC
    dset_attrs.pop("uuid")
    signal = np.concatenate(dsets)
    log.info("  ✓ %s: ", dset_name)
    for start, end in pad_intervals(intervals, pad_before, pad_after):
        start_sample = max(0, int(start * sampling_rate))
        end_sample = min(signal.size, int(end * sampling_rate))
        song_timestamp = timestamp_to_datetime(
            parent_entry.attrs["timestamp"]
        ) + timedelta(seconds=start_sample / sampling_rate / 1000)
        new_entry = writer.create_entry(song_timestamp, **entry_attrs)
        new_dset = create_dataset(
            new_entry,
            posixpath.basename(dsets[0].name),
            signal[start_sample:end_sample],
            compression="gzip",
            **dset_attrs,
        )
        set_uuid(new_dset)
        log.info(
            "    - samples %d–%d (%.1f ms) -> %s%s",   # noqa: RUF001
            start_sample,
            end_sample,
            (end_sample - start_sample) / sampling_rate,
            writer.filename,
            new_dset.name,
        )


def main(argv=None):
    p = argparse.ArgumentParser(
        description="train the song detector on labeled recordings"
    )
    p.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {__version__})",
    )
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    sub = p.add_subparsers(title="subcommands")

    pp = sub.add_parser("init", help="generate a starting configuration file")
    pp.set_defaults(func=make_config)
    pp.add_argument(
        "--train",
        action="store_true",
        help="run training with the intialized config file",
    )
    pp.add_argument(
        "config", type=Path, help="path where the configuration file should be saved"
    )
    pp.add_argument(
        "label_files",
        type=Path,
        nargs="*",
        help="label files to add to config file as training dataset",
    )

    pp = sub.add_parser("train", help="train the classifier")
    pp.set_defaults(func=train_classifier)
    pp.add_argument(
        "--dataset-dir",
        "-d",
        type=Path,
        default=Path("."),
        help="directory where training data files are stored",
    )
    pp.add_argument("config", type=Path, help="path of the configuration file (will be generated with default values if it doesn't exist)")
    pp.add_argument(
        "model",
        type=Path,
        nargs="?",
        help="path where the model should be saved (default is `config.pkl`)",
    )

    pp = sub.add_parser(
        "extract", help="use a trained model to extract songs from ARF files"
    )
    pp.set_defaults(func=extract_songs)
    pp.add_argument(
        "--dataset-name",
        default="pcm_000",
        help="the name of the dataset to process in each entry (default `%(default)s`)",
    )
    pp.add_argument(
        "--max-gap",
        type=float,
        default=100,
        help="merge segments separated by gaps less than this value (default %(default).1f ms)",
    )
    pp.add_argument(
        "--min-duration",
        type=float,
        default=500,
        help="segments must be at least this long (default %(default).1f ms)",
    )
    pp.add_argument(
        "--pad-before",
        type=float,
        default=1000,
        help="the amount of time before song onset to extract (default %(default).1f ms)",
    )
    pp.add_argument(
        "--pad-after",
        type=float,
        default=1000,
        help="the amount of time after song end to extract (default %(default).1f ms)",
    )
    pp.add_argument("--ignore-before", type=valid_datetime, help="ignore source entries before this time")
    pp.add_argument("--ignore-after", type=valid_datetime, help="ignore source entries after this time")
    pp.add_argument("model", help="trained classifier model (pkl file saved by `train`")
    pp.add_argument("input", help="the ARF file to process", nargs="+")
    pp.add_argument("output", help="output where the song segments should be saved")

    args = p.parse_args(argv)

    setup_log(log, args.debug)
    if not hasattr(args, "func"):
        p.print_usage()
        return 0

    args.func(args)


if __name__ == "__main__":
    main()



# -*- mode: python -*-
import logging
import unittest

import numpy as np

from quicksong import core
from quicksong.streaming import (
    STFT,
    IncompatibleBlockSize,
    OverlapSaveConvolver,
)

log = logging.getLogger("quicksong")


class TestConvolution(unittest.TestCase):

    signal = np.random.randn(5000)

    def test_bad_block_size(self):
        window_sizes = (11, 21, 51)
        kernels = core.make_hanning_kernels(window_sizes)
        convolver = OverlapSaveConvolver(kernels)
        dummy = np.zeros(convolver.block_size * 2)
        with self.assertRaises(IncompatibleBlockSize):
            convolver.process(dummy)

    def test_convolve(self):
        """Output of running convolution is the same as np.convolve"""
        window_sizes = (11, 20, 51)
        kernels = core.make_hanning_kernels(window_sizes)
        nt, nk = kernels.shape
        self.assertEqual(nt, max(window_sizes))
        self.assertEqual(nk, len(window_sizes))

        convolver = OverlapSaveConvolver(kernels)
        out = []
        convolver.prefill(self.signal[: convolver.block_size])
        for i in range(convolver.block_size, self.signal.size, convolver.block_size):
            s = self.signal[i : i + convolver.block_size]
            out.append(convolver.process(s))
        out.append(convolver.close())
        out = np.vstack(out)
        nout_t, nout_k = out.shape
        self.assertEqual(nout_k, nk)
        self.assertEqual(nout_t, self.signal.size)
        for i in range(nk):
            y = np.convolve(self.signal, kernels[:, i], mode="same")
            self.assertTrue(np.allclose(out[:, i], y))

    def test_copy_convolver(self):
        from copy import copy

        window_sizes = (11, 21, 51)
        kernels = core.make_hanning_kernels(window_sizes)
        conv1 = OverlapSaveConvolver(kernels)
        conv2 = copy(conv1)
        self.assertTrue(np.allclose(conv1.kernels, conv2.kernels))
        self.assertAlmostEqual(conv2.buffer.sum(), 0.0)


class TestSTFT(unittest.TestCase):

    sampling_rate = 48000
    frequency_range = (500, 20000)
    signal = np.random.randn(50000)

    def test_block_too_big(self):
        speccr = STFT(self.sampling_rate, 0.0020, 0.0005, self.frequency_range)
        dummy = np.zeros(speccr.block_size * 2)
        with self.assertRaises(IncompatibleBlockSize):
            speccr.process(dummy)

    def test_transform(self):
        """Output of running convolution is the same as np.convolve"""
        from math import ceil

        out = []
        speccr = STFT(self.sampling_rate, 0.0020, 0.0005, self.frequency_range)
        speccr.prefill(self.signal[: speccr.prefill_size])
        for i in range(speccr.prefill_size, self.signal.size, speccr.block_size):
            s = self.signal[i : i + speccr.block_size]
            out.append(speccr.process(s))
        out = np.column_stack(out)
        nf, nt = out.shape
        self.assertEqual(nf, speccr.findx.size)
        expected = ceil((self.signal.size - speccr.prefill_size) / speccr.block_size)
        self.assertEqual(nt, expected)





# -*- mode: python -*-
import logging
import unittest
from math import ceil

import numpy as np

from quicksong import core
from quicksong.features import total_power, wiener_entropy
from quicksong.streaming import STFT, OverlapSaveConvolver

log = logging.getLogger("quicksong")


class TestFeatureExtraction(unittest.TestCase):
    window_sizes = (11, 21, 51)
    sampling_rate = 48000
    window_ms = 20.0
    shift_ms = 10.0
    frequency_range = (500, 20000)
    signal = np.random.randn(20000)

    def make_extractor(self):
        speccr = STFT(
            self.sampling_rate,
            self.window_ms / 1000,
            self.shift_ms / 1000,
            self.frequency_range,
        )
        kernels = core.make_hanning_kernels(self.window_sizes)
        convolver = OverlapSaveConvolver(kernels)
        features = (total_power, wiener_entropy)
        return core.FeatureExtractor(speccr, features, convolver)

    def test_signal_duration(self):
        extractor_1 = self.make_extractor()
        out = []
        for block in extractor_1.process(self.signal):
            out.append(block)
        out.append(extractor_1.close())
        features_1 = np.vstack(out)
        expected = ceil(self.signal.size / extractor_1.frame_size)
        self.assertEqual(features_1.shape[0], expected)

    def test_signal_split(self):
        """output should be the same no matter how the input is divided up"""
        extractor_1 = self.make_extractor()
        out = []
        for block in extractor_1.process(self.signal):
            out.append(block)
        out.append(extractor_1.close())
        features_1 = np.vstack(out)

        extractor_2 = self.make_extractor()
        out = []
        chunk_size = self.signal.size // 5
        for i in range(0, self.signal.size, chunk_size):
            chunk = self.signal[i : i + chunk_size]
            for block in extractor_2.process(chunk):
                out.append(block)
        out.append(extractor_2.close())
        features_2 = np.vstack(out)

        self.assertSequenceEqual(features_1.shape, features_2.shape)
        self.assertTrue(np.allclose(features_1, features_2))


class TestIntervalFinder(unittest.TestCase):
    start = np.array(
        [
            59,
            101,
            392,
            479,
            532,
            820,
            836,
            989,
            1087,
            1276,
            1355,
            1670,
            1753,
            1921,
            2053,
            2174,
            2483,
            2711,
            2773,
            2847,
            2935,
            2943,
            3330,
            3639,
            4151,
        ]
    )
    lengths = np.array(
        [
            2,
            8,
            9,
            11,
            11,
            2,
            7,
            5,
            14,
            10,
            7,
            10,
            10,
            28,
            10,
            5,
            7,
            19,
            4,
            3,
            10,
            178,
            300,
            20,
            1,
        ]
    )
    max_gap = 100
    max_interval = 500
    expected = np.array([[2935, 3120], [3330, 3658]])

    def setUp(self):
        self.arr = np.ones(5000, dtype="i")
        for start, len in zip(self.start, self.lengths):
            self.arr[start : start + len] = 0

    def test_find_intervals(self):
        ifd = core.IntervalFinder(100, 500, 10)
        intervals = np.vstack(list(ifd.process(self.arr)))
        self.assertEqual(intervals.shape, self.expected.shape)
        log.debug("expected: %s\n, got: %s", self.expected, intervals)
        self.assertTrue(np.allclose(intervals, self.expected))

    def test_find_intervals_discontinuous(self):
        ifd = core.IntervalFinder(100, 500, 10)
        ivals = []
        block_size = 1000
        for i in range(0, self.arr.size, block_size):
            block = self.arr[i : i + block_size]
            for interval in ifd.process(block):
                ivals.append(interval)
        intervals = np.vstack(ivals)
        self.assertEqual(intervals.shape, self.expected.shape)
        log.debug("expected: %s\n, got: %s", self.expected, intervals)
        self.assertTrue(np.allclose(intervals, self.expected))


class TestPadIntervals(unittest.TestCase):
    def test_pad_single_interval(self):
        intervals = [(1000, 2000)]
        self.assertSequenceEqual(
            list(core.pad_intervals(intervals, 1000, 500)), [(0, 2500)]
        )

    def test_merge_overlapping_intervals(self):
        intervals = [(1000, 2000), (2500, 3000)]
        self.assertSequenceEqual(
            list(core.pad_intervals(intervals, 1000, 500)), [(0, 3500)]
        )

    def test_dont_merge_nonoverlapping_intervals(self):
        intervals = [(1000, 2000), (2500, 3000)]
        self.assertSequenceEqual(
            list(core.pad_intervals(intervals, 100, 200)), [(900, 2200), (2400, 3200)]
        )



# -*- mode: python -*-
import unittest

import numpy as np

from quicksong import features


def total_power(frame: np.ndarray) -> float:
    psd = np.abs(frame) ** 2
    return np.log10(np.sum(psd))


def wiener_entropy(frame: np.ndarray) -> float:
    psd = np.abs(frame) ** 2
    return np.log10(np.exp(np.mean(np.log(psd))) / np.mean(psd))


class TestCythonImpl(unittest.TestCase):
    def test_total_power(self):
        input = np.random.randn(1000) + np.random.randn(1000) * 1j
        self.assertTrue(np.allclose(total_power(input), features.total_power(input)))

    def test_wiener_entropy(self):
        input = np.random.randn(1000) + np.random.randn(1000) * 1j
        self.assertTrue(
            np.allclose(wiener_entropy(input), features.wiener_entropy(input))
        )
